!Copyright 2018 Rene Halver, Forschungszentrum Juelich GmbH, Germany
!Copyright 2018 Godehard Sutmann, Forschungszentrum Juelich GmbH, Germany
!
!Redistribution and use in source and binary forms, with or without modification, are 
!permitted provided that the following conditions are met:
!
!1. Redistributions of source code must retain the above copyright notice, this list 
!   of conditions and the following disclaimer.
!
!2. Redistributions in binary form must reproduce the above copyright notice, this 
!   list of conditions and the following disclaimer in the documentation and/or 
!   other materials provided with the distribution.
!
!3. Neither the name of the copyright holder nor the names of its contributors 
!   may be used to endorse or promote products derived from this software without 
!   specific prior written permission.
!
!THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
!EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
!OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
!SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
!INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
!TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
!BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
!CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
!ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
!SUCH DAMAGE.

! module for ALL access from Fortran
MODULE ALL_module
    USE ISO_C_BINDING
    IMPLICIT NONE
    PUBLIC
        TYPE ALL_t
            PRIVATE
            TYPE(c_ptr) :: object = C_NULL_PTR
        END TYPE
    ! definitions of enum type for different methods
        INTEGER(c_int), parameter ::  ALL_STAGGERED = 0
        INTEGER(c_int), parameter ::  ALL_TENSOR = 1
        INTEGER(c_int), parameter ::  ALL_UNSTRUCTURED = 2
        INTEGER(c_int), parameter ::  ALL_VORONOI = 3 
        INTEGER(c_int), parameter ::  ALL_HISTOGRAM = 4
    ! interface functions / subroutines to C++
    INTERFACE
        FUNCTION ALL_init_int(method, dim, gamma) RESULT(this) BIND(C,NAME="all_init_f")
            USE ISO_C_BINDING
            INTEGER(c_int),VALUE    :: method
            INTEGER(c_int),VALUE    :: dim
            REAL(c_double),VALUE    :: gamma
            TYPE(c_ptr)             :: this
        END FUNCTION
        SUBROUTINE ALL_set_proc_grid_params_int(obj, nloc, loc, nsize, size) BIND(C,NAME="all_set_proc_grid_params_f")
            USE ISO_C_BINDING
            TYPE(c_ptr), VALUE              ::  obj
            INTEGER(c_int), VALUE           ::  nloc
            INTEGER(c_int), DIMENSION(nloc) ::  loc
            INTEGER(c_int), VALUE           ::  nsize
            INTEGER(c_int), DIMENSION(nloc) ::  size
        END SUBROUTINE
        SUBROUTINE ALL_set_proc_tag_int(obj, tag) BIND(C,NAME="all_set_proc_tag_f")
            USE ISO_C_BINDING
            TYPE(c_ptr), VALUE              ::  obj
            INTEGER(c_int), VALUE           ::  tag
        END SUBROUTINE
        SUBROUTINE ALL_set_min_domain_size_int(obj, dim, domain_size) BIND(C,NAME="all_set_min_domain_size_f")
            USE ISO_C_BINDING
            INTEGER(c_int), VALUE           ::  dim
            REAL(c_double), DIMENSION(dim)  ::  domain_size
            TYPE(c_ptr), VALUE              ::  obj
        END SUBROUTINE
        SUBROUTINE ALL_set_work_int(obj, work) BIND(C,NAME="all_set_work_f")
            USE ISO_C_BINDING
            REAL(c_double),VALUE    ::  work
            TYPE(c_ptr), VALUE      ::  obj
        END SUBROUTINE
        SUBROUTINE ALL_set_vertices_int(obj, n, dim, vertices) BIND(C,NAME="all_set_vertices_f")
            USE ISO_C_BINDING
            TYPE(c_ptr), VALUE                  ::  obj
            INTEGER(c_int), VALUE               ::  n
            INTEGER(c_int), VALUE               ::  dim
            REAL(c_double), DIMENSION(n*dim)    ::  vertices
        END SUBROUTINE
        SUBROUTINE ALL_set_communicator_int(obj, comm) BIND(C,NAME="all_set_communicator_f")
            USE ISO_C_BINDING
            TYPE(c_ptr), VALUE              ::  obj
            INTEGER, VALUE                  ::  comm
        END SUBROUTINE
        SUBROUTINE ALL_setup_int(obj) BIND(C,NAME="all_setup_f")
            USE ISO_C_BINDING
            TYPE(c_ptr), VALUE              ::  obj
        END SUBROUTINE
        SUBROUTINE ALL_balance_int(obj) BIND(C,NAME="all_balance_f")
            USE ISO_C_BINDING
            TYPE(c_ptr), VALUE              ::  obj
        END SUBROUTINE
        SUBROUTINE ALL_get_new_number_of_vertices_int(obj,n) &
            BIND(C,NAME="all_get_new_number_of_vertices_f")
            USE ISO_C_BINDING
            TYPE(c_ptr), VALUE              ::  obj
            INTEGER(c_int)                  ::  n
        END SUBROUTINE
        SUBROUTINE ALL_get_vertices_int(obj,n,vertices) &
            BIND(C,NAME="all_get_vertices_f")
            USE ISO_C_BINDING
            TYPE(c_ptr), VALUE              ::  obj
            INTEGER(c_int), VALUE           ::  n
            REAL(c_double)                  ::  vertices(*) 
        END SUBROUTINE
        SUBROUTINE ALL_get_prev_vertices_int(obj,n,vertices) &
            BIND(C,NAME="all_get_prev_vertices_f")
            USE ISO_C_BINDING
            TYPE(c_ptr), VALUE              ::  obj
            INTEGER(c_int), VALUE           ::  n
            REAL(c_double)                  ::  vertices(*) 
        END SUBROUTINE
        SUBROUTINE ALL_print_vtk_outlines_int(obj, step) BIND(C,NAME="all_print_vtk_outlines_f")
            USE ISO_C_BINDING
            TYPE(c_ptr), VALUE              ::  obj
            INTEGER(c_int), VALUE           ::  step
        END SUBROUTINE
    END INTERFACE
    ! module subroutines to be called from Fortran code
    CONTAINS
        SUBROUTINE ALL_init(obj, method, dim, gamma)
            TYPE(ALL_t)     :: obj
            INTEGER         :: method
            INTEGER         :: dim
            REAL(c_double)  :: gamma
            obj%object = ALL_init_int(INT(method,c_int),INT(dim,c_int),REAL(gamma,c_double))
        END SUBROUTINE
        SUBROUTINE ALL_set_proc_grid_params(obj, loc, ranks)
            TYPE(ALL_t)                   ::  obj
            INTEGER, DIMENSION(:)         ::  loc
            INTEGER, DIMENSION(size(loc)) ::  ranks
            call ALL_set_proc_grid_params_int(obj%object, size(loc,1), loc, size(ranks,1), ranks)
        END SUBROUTINE
        SUBROUTINE ALL_set_proc_tag(obj, tag)
            TYPE(ALL_t)                   ::  obj
            INTEGER                       ::  tag
            call ALL_set_proc_tag_int(obj%object, tag)
        END SUBROUTINE
        SUBROUTINE ALL_set_min_domain_size(obj, dim, domain_size)
            INTEGER                         ::  dim
            REAL(c_double), DIMENSION(dim)  ::  domain_size
            TYPE(ALL_t)                     ::  obj
            CALL ALL_set_min_domain_size_int(obj%object, dim, domain_size)
        END SUBROUTINE
        SUBROUTINE ALL_set_work(obj, work)
            TYPE(ALL_t)     ::  obj
            REAL(c_double)  ::  work
            CALL ALL_set_work_int(obj%object, work)
        END SUBROUTINE
        SUBROUTINE ALL_set_vertices(obj, n, dim, vertices)
            TYPE(ALL_t)     ::  obj
            INTEGER         ::  n
            INTEGER         ::  dim
            REAL(c_double)  ::  vertices(n*dim)
            CALL ALL_set_vertices_int(obj%object, n, dim, vertices)
        END SUBROUTINE
        SUBROUTINE ALL_set_communicator(obj, comm)
            TYPE(ALL_t)     ::  obj
            INTEGER         ::  comm
            CALL ALL_set_communicator_int(obj%object, comm)
        END SUBROUTINE
        SUBROUTINE ALL_setup(obj)
            USE ISO_C_BINDING
            TYPE(ALL_t)         ::  obj
            CALL ALL_setup_int(obj%object)
        END SUBROUTINE
        SUBROUTINE ALL_balance(obj)
            USE ISO_C_BINDING
            TYPE(ALL_t)         ::  obj
            CALL ALL_balance_int(obj%object)
        END SUBROUTINE
        SUBROUTINE ALL_get_new_number_of_vertices(obj,n)
            TYPE(ALL_t)         ::  obj
            INTEGER             ::  n
            CALL ALL_get_new_number_of_vertices_int(obj%object,n)
        END SUBROUTINE
        SUBROUTINE ALL_get_vertices(obj,n,vertices)
            TYPE(ALL_t)         ::  obj
            INTEGER             ::  n
            REAL(c_double)      ::  vertices(*)
            CALL ALL_get_vertices_int(obj%object,n,vertices)
        END SUBROUTINE
        SUBROUTINE ALL_get_prev_vertices(obj,n,vertices)
            TYPE(ALL_t)         ::  obj
            INTEGER             ::  n
            REAL(c_double)      ::  vertices(*)
            CALL ALL_get_prev_vertices_int(obj%object,n,vertices)
        END SUBROUTINE
        SUBROUTINE ALL_print_vtk_outlines(obj,step)
            TYPE(ALL_t)    ::  obj
            INTEGER        ::  step
            CALL ALL_print_vtk_outlines_int(obj%object,step)
        END SUBROUTINE
END MODULE ALL_module

